"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.removeCommonPrefix = exports.addLinks = exports.mergeTrees = exports.flattenTree = exports.buildTree = void 0;
const path_1 = __importDefault(require("path"));
const types_1 = require("../types/types");
const PLUGIN_PREFIX = "\u0000";
const addToPath = (tree, modulePath, node) => {
    if (modulePath.length === 0) {
        throw new Error(`Error adding node to path ${modulePath.join("/")}`);
    }
    const [head, ...rest] = modulePath;
    if (rest.length === 0) {
        tree.children.push({ ...node, name: head });
        return;
    }
    else {
        let newTree = tree.children.find((folder) => folder.name === head && types_1.isModuleTree(folder));
        if (!newTree) {
            newTree = { name: head, children: [] };
            tree.children.push(newTree);
        }
        addToPath(newTree, rest, node);
        return;
    }
};
const buildTree = (modules, mapper) => {
    let tree = {
        name: "root",
        children: [],
    };
    for (const { id, renderedLength, gzipLength, brotliLength } of modules) {
        const mod = { id, renderedLength, gzipLength, brotliLength };
        const name = id;
        const uid = mapper.setValueByModuleId(id, mod);
        const nodeData = { uid };
        if (name.startsWith(PLUGIN_PREFIX)) {
            addToPath(tree, [name], nodeData);
        }
        else {
            const pathParts = name.split(path_1.default.sep);
            pathParts.shift(); // remove '' or C:
            addToPath(tree, pathParts, nodeData);
        }
    }
    tree = exports.flattenTree(tree);
    return tree;
};
exports.buildTree = buildTree;
// if root children have only on child we can flatten this
const flattenTree = (root) => {
    const pluginChildren = [];
    const otherChildren = [];
    for (const child of root.children) {
        if (child === null || child === void 0 ? void 0 : child.name.startsWith(PLUGIN_PREFIX)) {
            pluginChildren.push(child);
        }
        else {
            otherChildren.push(child);
        }
    }
    let newRoot = root;
    if (otherChildren.length === 1 && types_1.isModuleTree(otherChildren[0])) {
        newRoot = otherChildren[0];
    }
    while (newRoot.children.length === 1 && types_1.isModuleTree(newRoot.children[0])) {
        newRoot = newRoot.children[0];
    }
    newRoot.children = newRoot.children.concat(pluginChildren);
    return newRoot;
};
exports.flattenTree = flattenTree;
const mergeTrees = (trees) => {
    if (trees.length === 1 && types_1.isModuleTree(trees[0])) {
        return trees[0];
    }
    const newTree = {
        name: "root",
        children: trees,
    };
    return newTree;
};
exports.mergeTrees = mergeTrees;
const addLinks = (startModuleId, getModuleInfo, links, mapper) => {
    const processedNodes = {};
    const moduleIds = [startModuleId];
    while (moduleIds.length > 0) {
        const moduleId = moduleIds.shift();
        const moduleUid = mapper.getUid(moduleId);
        if (processedNodes[moduleUid]) {
            continue;
        }
        else {
            processedNodes[moduleUid] = true;
        }
        const mod = mapper.getValue(moduleUid, { id: moduleId, renderedLength: 0 });
        const moduleInfo = getModuleInfo(moduleId);
        if (!moduleInfo) {
            return;
        }
        if (moduleInfo.isEntry) {
            mod.isEntry = true;
        }
        if (moduleInfo.isExternal) {
            mod.isExternal = true;
        }
        for (const importedId of moduleInfo.importedIds) {
            const importedUid = mapper.getUid(importedId);
            links.push({ source: moduleUid, target: importedUid });
            moduleIds.push(importedId);
        }
        for (const importedId of moduleInfo.dynamicallyImportedIds) {
            const importedUid = mapper.getUid(importedId);
            links.push({ source: moduleUid, target: importedUid, dynamic: true });
            moduleIds.push(importedId);
        }
    }
};
exports.addLinks = addLinks;
const skipModule = (id, node) => id.startsWith(PLUGIN_PREFIX) || node.isExternal || !path_1.default.isAbsolute(id);
const removeCommonPrefix = (nodes, nodeIds) => {
    let commonPrefix = null;
    for (const [id, uid] of Object.entries(nodeIds)) {
        const node = nodes[uid];
        if (!skipModule(id, node)) {
            if (commonPrefix == null) {
                commonPrefix = id;
            }
            for (let i = 0; i < commonPrefix.length && i < id.length; i++) {
                if (commonPrefix[i] !== id[i]) {
                    commonPrefix = commonPrefix.slice(0, i);
                    break;
                }
            }
        }
    }
    if (commonPrefix == null) {
        return;
    }
    const commonPrefixLength = commonPrefix.length;
    for (const [id, uid] of Object.entries(nodeIds)) {
        const node = nodes[uid];
        if (!skipModule(id, node)) {
            const newId = id.slice(commonPrefixLength);
            const value = nodeIds[id];
            delete nodeIds[id];
            nodeIds[newId] = value;
        }
    }
};
exports.removeCommonPrefix = removeCommonPrefix;
